/* RUNTEST.C - contains the runtime code needed to read and parse
               function test scripts.

   <b>runtest.c</b> contains all the code needed to carry out
   unit testing on one of more C function calls. Unit tests are
   defined within test scripts. These are plain text files  which
   contain a sequence of comment lines used to describe tests and their
   expected results along with, blank lines for readability and function
   test definitions. Each test definition is a line containing the name
   of the function to be executed, a list of parameter values and an
   optional comment at the end of the line, typically used to say whether
   this test is expected to succeed or fail and to summarise its expected
   return value, if any.

   Comments start with a '#' and are terminated with the end of the line.

   Test definitions must start with the name of the function being tested.
   this is followed by up to 10 parameter values separated from the function
   name and from each other by at least one whitespace character. An optional
   comment may follow the last parameter.

   Parameters may be unquoted, in which case they must not include whitespace.
   Quoted parameters may include whitespace. The quotemarks are either a
   single quote (') or a double quote (") but the endmark must be the
   same as the opening quotemark. Inside a quoted parameter to 'other'
   quotemark looses its property of being an endquote, so both
   "quoted 'string'" and  'another quoted "string"' will be interpreted
   correctly, while "quoted "string"" will be treated as two parameters:
   'quoted ' and 'string""'.

   A short example script might be: 

   ----------------------------------------------------------------
   # re01.txt - clean re_error() test.
   # ======== - error with simple text description
   #

   reporterror re01                # OK
   re_error "error text 1"         # "re01: error text 1" expected

   #
   # No errors expected
   #
   ----------------------------------------------------------------

   
   

            Written by M.C.Gregorie.

            Version history
               
            1.0   26/05/20   Initial release
            2.0   31/07/20   Renamed testharness() to runtest()
*/

#define VERSION         "2.0"
#define COPYRIGHT       "(c) M.C.Gregorie, 2020"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "runtest.h"
#include "custom.h"

extern int  errno;

int         errors = 0;          /* errors may be internal or a test result */
char        *progname;           /* program name on the command line        */
char        *script[MAXSCRIPTS]; /* lists of scripts to run                 */
int         scriptno = 0;        /* size of script list                     */

/*
   error - reports errors in thetest harness itself
   =====   and terminates the run
*/
void error(char *desc, char *value)
{
   fprintf(stderr, "%s Error: %s %s\n", progname, desc, value);
   exit(1);
}

/*
   This function scans the command line for options and does
   all the associated setup.
*/
void readargs(int argc, char *argv[], int flags[])
{
   int         i;
   int         j;
   char        opt[2];

   /* initialise flags */

   for (i = 0; i < FLAGCNT; i++)
      flags[i] = FALSE;

   /* set up options from the command line */

   for (i = 0; i < argc; i++)   
   {
      if (i == 0)
         progname = argv[0];
      else
      if (argv[i][0] == '-')
      {
         /*
            Found an option or a set of concatenated options.
         */
         for (j = 1; j < strlen(argv[i]); j++)
         {
            switch (argv[i][j])
            {
               case '?': flags[HELP] = TRUE;       
                         break;
               case 'd': flags[DEBUG] = flags[DEBUG] + 1;
                         break;
               default : opt[0] = argv[i][j];
                         opt[1] = 0;
                         error("invalid option:", opt);
            }
         }
      }
      else
      {
         /*
            Not an option, so treat this argument as a script file name.
         */
         if (scriptno >= MAXSCRIPTS)
            error("too many scripts specified", "");
         else
            script[scriptno++] = argv[i];   
      }
   }
}

/*
   runtest - processes every line in a test script.
   =======
   Comment lines may appear anywhere and have a hash, #, as thre first
   character. Comments and blank lines are copied to output but otherwise
   ignored.
   
   Each nom-comment line defines a function call which takes the form

   function param1 param2 .... # comment documenting expected results

   Where the function name must start at the first character in the line
   and parameters are separated by one or more spaces. There may be up to
   10 parameters. Quoted parameters may contain spaces: unquoted parameters
   most not contain spaces.

   If a parameter contains spaces it must be 'quoted' by enclosing it in
   in single or double quotes ( 'param text' or "param text" ). 
   Quoted parameters must be terminated with the same single or double
   quote mark that they started with, but may contain the other type of
   quote, so "String val '%s'" and 'Char val "%c"' will both
   be parsed correctly but "stringvalue' will include all following
   parameters until either a double quote or the end of the parameter list
   is found.

   The parameter list may be followed by a comment, starting with a hash,
   which as a useful way of documenting the expected results from a
   function call.

   When the test has been parsed into function name plus a list of parameter
   strings, a parallel list of numeric values is produced. Each parameter
   is first tested to see if it is a boolean value. 

   If its first character is T,t or 1 its numeric value is set to 1 (TRUE)
   and if its first chareacter is F,f or 0 its numeric value to 0 (FALSE).

   If neither comparison matches, the parameter is used as input to atoi()
   and rhe result is used as its numeric value.
*/
void runtest(FILE* f, int debug)
{
   char  c;                   /* c is used to classify lines: those
                                 starting with '#' '\n' or space are
                                 listed and ignored.
                              */ 
   int   i;                   /* i        is the line index                */
   int   j;                   /* j        is fn name and param write index */ 
   int   param;               /* param    is the parameter index           */
   int   state;               /* state    FN        for fn name,
                                          SEPSCAN for separator,
                                          NONQUOTE  for unquoted parameter
                                          QUOTED    for quoted parameter   */
   int   limit;               /* limit    is the line length               */
   int   lineno = 0;          /* lineno   is the current line number       */
   char  line[MAXLINE + 1];   /* line     the current line of a test defn  */
   char  fn[MAXLINE];         /* fn       is the name of the function to be
                                          be tested                        */
   char  s[MAXARGS][MAXLINE]; /* s        contains string parameters for
                                          the test.                        */
   int   n[MAXARGS];          /* n        contains integer parameters for  
                                          the test.
                                          Set to zero if the matching
                                          string is not numeric.            */
   int   reply = 0;           /* reply    is the return value from
                                          calladapter()                     */
   char  terminator;          /* terminator contains the character that began
                                          a quoted parameter and that will
                                          end it. This makes constructs like
                                          " param = '%s'" or ' val = "%d"'
                                          parse correctly                   */
                                           

   char  x = 0;               /* Used in boolean value tests                */

   errors = 0;
   while (fgets(line, MAXLINE, f) > 0)
   {
      /* Write line no and line to stdout */
      lineno++;
      printf("%4d: %s", lineno, line);
      c = line[0];

      /* Ignore line if it starts with '#', space or newline */
      if (c == '#' || c == ' ' || c == '\n') 
         continue;                     

      /*
         Scan the line, looking for either a trailing comment or the
         final newline. The first found is replaced by a NULL so to  
         terminate the following scan for the function name and parameters. 
      */
      limit = strlen(line);
      for (i = 0; i < limit; i++)
         if (line[i] == '#' || line[i] == '\n')
         {
            line[i] = 0;
            limit = i;
            break;
         }

      if (debug > 1)
         printf("\n      test command: '%s'\n\n", line);

      /*
         Extract fn name and parameters from the line.
      */
      state = FN;
      param = 0;
      j = 0;
      for (i = 0; i < limit && i < MAXLINE && param < MAXARGS; i++)
      {
         if (debug > 2)
            fprintf(stderr, "'%c' i=%d state=%d param=%d j=%d\n",
                            line[i], i, state, param, j);
            
         switch (state)
         {
            case FN:       if (i < limit && !isspace(line[i]))
                           {
                              fn[j++] = line[i];
                              fn[j] = 0;
                           }
                           else
                           {
                              state = SEPSCAN;
                              j = 0;
                           }

                           break;
                           
            case SEPSCAN:  if (line[i] == '"' || line[i] == 0x27)
                           {
                              state = QUOTED;
                              terminator = line[i];
                           }
                           else
                           if (!isspace(line[i]))
                           {
                              state = UNQUOTED;
                              s[param][j++] = line[i];
                           }
                           
                           break;
            
            case QUOTED:   if (line[i] == terminator)
                           {
                              state = SEPSCAN;
                              s[param][j] = 0;
                              j = 0;
                              param++;
                           }
                           else
                              s[param][j++] = line[i];

                           break;
                                 
            case UNQUOTED: if (isspace(line[i]) ||
                               line[i] == '\n' ||
                               line[i] == 0)
                           {
                              state = SEPSCAN;
                              s[param][j] = 0;
                              j = 0;
                              param++;
                           }
                           else
                              s[param][j++] = line[i];

                           break;                                 
                                 
            default:       fprintf(stderr,
                                   "%s: %s at line %d char %d: '%c'\n",
                                   progname,
                                   "Unknown state",
                                   lineno,
                                   i,
                                   line[i]);
         }   
      }

      if (state == UNQUOTED)
      {
         /*
            no whitespace or end quote after the last param
         */
         s[param][j] = 0;
         param++;
      }

      /*
         Now try to convert string parameters into integers.
      */
      if (debug > 1)
         fprintf(stderr, "Setting integers\n");
      
      for (j = 0; j < MAXARGS; j++)
      {
         if (j < param)
         {
            x = s[j][0];
            if (x == 'T' || x == 't' || x =='1')
               n[j] = 1;
            else
            if (x == 'F' || x == 'f' || x =='0')
               n[j] = 0;
            else
               n[j] = atoi(s[j]);
         }
         else
         {
            s[j][0] = 0;
            n[j] = 0;
         }
      }

      if (debug)
      {
         fprintf(stderr, "%s:\n", fn);
         fprintf(stderr, "  : ");
         for (i = 0; i < param; i++)
            fprintf(stderr, "'%s' " , s[i]);
            
         fprintf(stderr, "\n  : ");
         for (i = 0; i < param; i++)
            fprintf(stderr, "%d " , n[i]);

          fprintf(stderr, "\n");
      }

      reply = calladapter(debug, fn, s, n);
      if (reply)
      {
         fprintf(stderr, "Run abandoned by calladapter() (reply=%d)\n",
                         reply);
         exit(reply);
      }  
   }

   fclose(f);
   if (errors == 0)
      printf("\nNo errors found\n\n");
   else
   if (errors == 1)
      printf("\n1 error found\n\n");
   else
      printf("\n%d errors found\n\n", errors);
}

/*
   Report a true/false return value
*/
void setboolreply(int n)
{
   printf("      Returned: %s\n\n", (n == 0 ? "false" : "true"));
}

/*
   Report a character return value
*/
void setcharreply(char c)
{
   printf("      Returned: '%c'\n\n", c);
}

/*
   Report a double return value
*/
void setdoublereply(double d)
{
   printf("      Returned: %f\n\n", d);
}

/*
   Report an error
*/
void seterror(char *e)
{
   printf("      Error: %s\n\n", e);
   errors++;
}

/*
   Report a float return value
*/
void setfloatreply(float f)
{
   printf("      Returned: %f\n\n", f);
}

/*
   Report an integer return value
*/
void setintreply(int r)
{
   printf("      Returned: %d\n\n", r);
}

/*
   Report a string return value
*/
void setstringreply(char *s)
{
   printf("      Returned: %c%s%c\n\n", '"', s, '"');
}

/*
   The main program starts here
   ----------------------------
*/
int main(int argc, char *argv[])
{
   FILE  *f;
   int   flags[FLAGCNT];
   int   i;

   /*
      Parse and set up command options
   */
   readargs(argc,argv,flags);

   /*
      If help was asked for, give it and quit
   */
   if (flags[HELP]) 
   {
      showhelp();
      return(0);
   }

   /*
      The following code drives the application. 
   */
   i = 0;
   while (i < scriptno)
   {
      if (flags[DEBUG])
      {
         printf("Processing file %s\n", script[i]);
      }

      f = fopen(script[i], "r");
      if (f == NULL)
         error("can't open", script[i]);
      else
         runtest(f, flags[DEBUG]);

      i++;
   }

   return(0);
}
